// General-purpose Photovoltaic Device Model gpvdm.com - a drift diffusion
// base/Shockley-Read-Hall model for 1st, 2nd and 3rd generation solarcells.
// The model can simulate OLEDs, Perovskite cells, and OFETs.
// 
// Copyright (C) 2008-2020 Roderick C. I. MacKenzie
// All rights reserved.
// 
// https://www.gpvdm.com
// r.c.i.mackenzie at googlemail.com
// 
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the GPVDM nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL Roderick C. I. MacKenzie BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "sim.h"
#include <math.h>

double sim_ce(struct device *in)
{
int i;
set_ntricks_fast(TRUE);
in->dt=in->cedt;
char outpath[1000];
double extractedcharges=0.0;
FILE *train;
FILE *iv;

carrier_count_reset(in);

in->go_time=FALSE;

#ifndef _fast_
train=fopena(in->outputpath,"./cefit.dat","w");
fclose(train);

train=fopena(in->outputpath,"./ce.dat","w");
fclose(train);

train=fopena(in->outputpath,"./gex.dat","w");
fclose(train);

iv=fopena(in->outputpath,"./cen.dat","w");
fclose(iv);
#endif


printf("Charge Extraction V= %le Sun=%le\n",in->cevoltage,in->Psun);

int ittr=0;
int step=0;
char out[1000];

struct istruct cejn;
inter_init(&cejn);

struct istruct cejp;
inter_init(&cejp);

struct istruct cetrap;
inter_init(&cetrap);

struct istruct cer;
inter_init(&cer);

struct istruct ceextracted;
inter_init(&ceextracted);

struct istruct cei;
inter_init(&cei);

double Psun_orig=in->Psun;
light_field(in,in->spectra,0.0);
in->Vapplied=in->cevoltage;
solve_all(in);

printf("Vapplied=%le\n",in->Vapplied);
//ramp(in,in->Vapplied,in->cevoltage,0.1);
newton_set_min_ittr(100);
sim_externalv(in,in->cevoltage);
newton_set_min_ittr(0);

//sim_externalv(in,in->cevoltage);
//solve_all(in);
plot_now(in,"plot");


device_timestep(in);

in->go_time=TRUE;

int turnofflight=TRUE;

in->time=-in->cepresteps*in->dt;
double n_start=0.0;
double p_start=0.0;
double np_out_tot=0.0;
double r_tot=0.0;
double last_n=0.0;
int cestep=0;
do
{
if (step==in->cepresteps) carrier_count_reset(in);

	if (step<in->cepresteps)
	{
		sim_externalv(in,in->cevoltage);
		//printf("external %lf\n",in->cevoltage);
	}else
	{
		if (turnofflight==TRUE)
		{
			n_start=get_extracted_n(in);
			p_start=get_extracted_p(in);
			//printf("%le %le \n",(n_start+p_start)/2.0,get_extracted_np(in));
			//getchar();
			in->Psun=0;
			light_field(in,in->spectra,0.0);
			extractedcharges=0;
			np_out_tot=0.0;
			r_tot=0.0;
			turnofflight=FALSE;
		}

		newton_charge_extraction(in);

		cestep++;
	}
	if (cestep>10) set_ntricks_fast(TRUE);
//getchar();
	#ifndef _fast_
	printf("ce time=%e step=%d %.1e ",in->time,step,in->last_error);
	sprintf(out,"V=%lf current = %e mA (%e A/m^2)\n",in->Vapplied,get_I(in)/1e-3,get_J(in));

	printdebug(out);
	#endif

	ittr++;

	get_avg_recom(in);

	#ifndef _fast_
	plot_now(in,"plot");

	//("%le %le %le\n",(n_start+p_start)/2.0,get_extracted_np(in),fabs(np_out_tot)/2.0);

	iv=fopena(in->outputpath,"./cen.dat","a");
	fprintf(iv,"%le %le %le %le %le\n",in->time,get_free_n_charge(in),get_free_p_charge(in),get_n_trapped_charge(in),get_p_trapped_charge(in));
	fclose(iv);

	iv=fopena(in->outputpath,"./cefit.dat","a");
	fprintf(iv,"%e %e\n",in->time,get_I_ce(in));
	fclose(iv);

	iv=fopena(in->outputpath,"./ce.dat","a");
	fprintf(iv,"%e %e %e\n",in->time,get_I_ce(in),in->Vapplied);
	fclose(iv);

	train=fopena(in->outputpath,"./gex.dat","a");
	fprintf(train,"%.20e %.20e\n",in->time,in->Gex[0]);
	fclose(train);
	#endif




	inter_append(&cejn,in->time,get_extracted_n(in));
	inter_append(&cejp,in->time,get_extracted_p(in));

	extractedcharges+=(get_I_ce(in)/q)*in->dt;
	np_out_tot+=2.0*get_J(in)*in->dt/in->ylen/q;
	r_tot+=get_avg_recom(in)*in->dt;
	if (in->time>0.0) in->dt*=in->cedtmull;

	if (in->time>0.0)
	{
		inter_append(&ceextracted,in->time,100.0*(fabs(np_out_tot)/(n_start+p_start)));
		inter_append(&cer,in->time,100.0*r_tot/((n_start+p_start)/2.0));
		inter_append(&cetrap,in->time,100.0*(get_extracted_np(in)*2.0)/(n_start+p_start));
		inter_append(&cei,in->time,get_I_ce(in));

	}

	if ((in->cefast==TRUE)&&(cestep>2))
	{
		if (fabs(last_n-np_out_tot)<1e6)
		{
			printf("ce converged\n");
			break;
		}
	}

	device_timestep(in);


	step++;
//getchar();

	last_n=np_out_tot;
	if (in->time>in->cetimestop) break;


	if (in->stop==TRUE)
	{
		break;
	}
if (in->stop_start==TRUE) getchar();
}while(1);

in->go_time=FALSE;
extractedcharges-=in->cevoltage*in->C/q;
extractedcharges/=in->Vol;
inter_save_a(&cejn,in->outputpath,"cejn.dat");
inter_free(&cejn);

inter_save_a(&cejp,in->outputpath,"cejp.dat");
inter_free(&cejp);


inter_save_a(&ceextracted,in->outputpath,"ceextracted.dat");
FILE *info;

info=fopen("./ceinfo.dat","w");
fprintf(info,"%le %le %le %le %le %le\n",Psun_orig,in->cevoltage,inter_get_hard(&ceextracted,10e-3),inter_get_hard(&cer,10e-3),inter_get_hard(&cetrap,10e-3),fabs(np_out_tot)/2.0);
fclose(info);

info=fopen("./cen_tot.dat","w");
fprintf(info,"%le %le\n",in->cevoltage,extractedcharges);
fclose(info);

inter_save_a(&cetrap,in->outputpath,"cetrap.dat");
inter_free(&cetrap);

inter_save_a(&cer,in->outputpath,"cer.dat");
inter_free(&cer);

inter_mod(&cei);
inter_save_a(&cei,in->outputpath,"cei.dat");

if (in->dumpitdos==TRUE)
{
	inter_chop(&cei,1e-9, 1.0);

	for (i=0;i<cei.len;i++)
	{
	cei.data[i]*=cei.x[i];
	cei.x[i]=(25e-3*log(2e7*cei.x[i]));
	}

	//inter_chop(&tpcfitout,1e-9, 1.0);

	double val=inter_get_hard(&cei,0.0);

	inter_rescale(&cei,1.0, 1.0/val);

	sprintf(outpath,"%s%s",in->outputpath,"cei_itdos.dat");
	inter_save(&cei,outpath);

}

inter_free(&cei);

inter_free(&ceextracted);

return extractedcharges;
}
//</clean>
