// General-purpose Photovoltaic Device Model gpvdm.com - a drift diffusion
// base/Shockley-Read-Hall model for 1st, 2nd and 3rd generation solarcells.
// The model can simulate OLEDs, Perovskite cells, and OFETs.
// 
// Copyright (C) 2008-2020 Roderick C. I. MacKenzie
// All rights reserved.
// 
// https://www.gpvdm.com
// r.c.i.mackenzie at googlemail.com
// 
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the GPVDM nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL Roderick C. I. MacKenzie BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "sim.h"
#include <math.h>

void sim_celiv(struct device *in)
{
in->dt=in->celiv_dt;

carrier_count_reset(in);

in->go_time=FALSE;


//struct istruct celivout;
//inter_init(&celivout);
in->externalv=in->celiv_Vexternal;
printf("Running celiv. %e\n",in->externalv);
int ittr=0;

int step=0;
double dj;
double j0;
light_field(in,in->spectra,0.0);
//in->Vapplied=0.1;
//ramp(in,in->Vapplied,in->celiv_Vmax,100.0);
//sim_externalv(in,in->celiv_Vmax);
//double celiv_max=in->celiv_Vmax;

//ramp(in,in->Vapplied,in->celiv_Vexternal,100.0);
sim_externalv(in,in->celiv_Vexternal);
in->Vapplied=in->celiv_Vexternal;
double V=in->Vapplied;
double V0=in->celiv_Vexternal;

light_update(in,0.0);
newton_celiv(in,V0,FALSE);
device_timestep(in);
newton_celiv(in,V0,FALSE);
device_timestep(in);

//printf("\n\n%le\n",in->Vapplied_last);
//exit(0);

//printf("Vapplied_last %le %le\n",in->Vapplied,in->Vapplied_last);

time_init(in);

device_timestep(in);

in->go_time=TRUE;


in->time=0.0;
//double np_out_tot=0.0;
double t0=in->celiv_t0;
double t1=in->celiv_t0+in->celiv_t1;
double t2=in->celiv_t0+in->celiv_t1+in->celiv_t2;
FILE *celiv_out=fopen("./celiv_j.dat","w");
fclose(celiv_out);

celiv_out=fopen("./celiv_v.dat","w");
fclose(celiv_out);

celiv_out=fopen("./celiv_nleft.dat","w");
fclose(celiv_out);

celiv_out=fopen("./celiv_rtot.dat","w");
fclose(celiv_out);

double ramp_rate=in->celiv_Vmax*(in->dt/(t1-t0));
double local_time=in->time;
double i_max=0.0;
double t_max=0.0;
double mu=0.0;
double A=fabs(in->celiv_Vmax/(t1-t0));
double r_tot=0.0;
double n_start=0.0;
double p_start=0.0;
struct istruct out_j;
inter_init(&out_j);

struct istruct out_v;
inter_init(&out_v);

struct istruct out_mue;
inter_init(&out_mue);

struct istruct out_muh;
inter_init(&out_muh);
//	printf("%le %le\n",in->Vapplied,in->Vapplied_last);
	//exit(0);
do
{


	if (local_time<t0)
	{
		reset_npinit(in);
		V=V0;

		r_tot=0.0;
	}else
	if (local_time<t1)
	{
		V+=ramp_rate;

		celiv_out=fopen("./celiv_nleft.dat","a");
		fprintf(celiv_out,"%le %le\n",in->time,get_extracted_np(in));
		fclose(celiv_out);

		//light_update(in,in->toflaserpower);
	}else
	{
		V=V0;
	}


	//solve_all(in);

	double i0=-newton_celiv(in,V,TRUE);

	if ((local_time>t0)&&(local_time<t1))
	{
		if (i0>i_max)
		{
			i_max=i0;
			t_max=local_time-t0;
		}

		n_start=fabs(get_extracted_n(in));
		p_start=fabs(get_extracted_p(in));
		r_tot+=get_avg_recom(in)*in->dt;

		celiv_out=fopen("./celiv_rtot.dat","a");
		fprintf(celiv_out,"%le %le\n",in->time,100.0*r_tot/((n_start+p_start)/2.0));
		fclose(celiv_out);
	}
	#ifndef _fast_
	printf("celiv time=%e step=%d %.1e ",in->time,step,in->last_error);
	char out[100];
	sprintf(out,"Vtot=%lf current = %e mA (%e A/m^2)\n",V,get_I(in)/1e-3,get_J(in));

	printdebug(out);
	#endif

	ittr++;

	plot_now(in,"plot");

	//inter_append(&celivout,in->time,fabs(get_I(in)));

	inter_append(&out_j,in->time,i0);
	inter_append(&out_v,in->time,V);
	inter_append(&out_mue,in->time,get_avg_mue(in));
	inter_append(&out_muh,in->time,get_avg_muh(in));

	device_timestep(in);
	step++;

	if (in->time>(((double)in->celiv_repeat)*t2)) break;
	if (local_time>t2) local_time=0.0;
	local_time+=in->dt;

	if (in->stop==TRUE)
	{
		break;

	}


}while(1);
char outpath[1000];
sprintf(outpath,"%s%s",in->outputpath,"celiv_j.dat");
inter_save(&out_j,outpath);


sprintf(outpath,"%s%s",in->outputpath,"celiv_v.dat");
inter_save(&out_v,outpath);

sprintf(outpath,"%s%s",in->outputpath,"celiv_mue.dat");
inter_save(&out_mue,outpath);

sprintf(outpath,"%s%s",in->outputpath,"celiv_muh.dat");
inter_save(&out_muh,outpath);


printf("celiv stats\n");

j0=fabs(inter_get(&out_j,in->celiv_time_j1)-inter_get(&out_j,in->celiv_time_j0));
dj=fabs(inter_get(&out_j,in->celiv_time_j1)-i_max);
mu=(2.0/3.0)*in->ylen*in->ylen/(A*t_max*t_max*(1.0+0.21*(dj/j0)));

printf("%le %le\n",in->ylen*in->ylen,(A*t_max*t_max*(1.0+0.21*(dj/j0))));

celiv_out=fopen("./celiv_mu_value.dat","w");
fprintf(celiv_out,"%le\n",mu);
fclose(celiv_out);

in->go_time=FALSE;

inter_free(&out_muh);
inter_free(&out_mue);
inter_free(&out_j);
inter_free(&out_v);

}
//</clean>
