// General-purpose Photovoltaic Device Model gpvdm.com - a drift diffusion
// base/Shockley-Read-Hall model for 1st, 2nd and 3rd generation solarcells.
// The model can simulate OLEDs, Perovskite cells, and OFETs.
// 
// Copyright (C) 2008-2020 Roderick C. I. MacKenzie
// All rights reserved.
// 
// https://www.gpvdm.com
// r.c.i.mackenzie at googlemail.com
// 
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the GPVDM nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL Roderick C. I. MacKenzie BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "sim.h"
#include <math.h>



void sim_fish(struct device *in)
{
int i;
double nx=0.0;
double px=0.0;
in->dt=in->fish_dt;

carrier_count_reset(in);

in->go_time=FALSE;

printf("Running fish. %e\n",in->fish_externalv);
int ittr=0;
int step=0;

light_field(in,in->spectra,0.0);

sim_externalv(in,in->fish_externalv);
time_init(in);

for (i=0;i<in->ymeshpoints;i++)
{
	if ((in->ymesh[i]>in->fish_1start)&&(in->ymesh[i]<in->fish_1stop))
	{
		in->p[i]+=in->fish_1den;
	}

	if ((in->ymesh[i]>in->fish_2start)&&(in->ymesh[i]<in->fish_2stop))
	{
		in->n[i]+=in->fish_2den;
	}
printf("%le %le %le\n",in->ymesh[i],in->n[i],in->p[i]);
}

device_timestep(in);

in->go_time=TRUE;

in->time=0.0;

struct map frame;
frame_init(&frame,20,100,in->ymesh[0],in->ymesh[in->ymeshpoints-1],-6.0,-4.0);
light_update(in,0.0);

	FILE *out=fopen("./fish.dat","w");
	fclose(out);

struct istruct fishoutn;
inter_init(&fishoutn);

struct istruct fishoutp;
inter_init(&fishoutp);

do
{

	frame_reset(&frame);
	frame_build(&frame,in);
	frame_dump(&frame,in);

	solve_all(in);

	ittr++;

	plot_now(in,"plot");

	if (in->stop_start==TRUE) getchar();

	in->dt*=in->fish_dt_mull;

	device_timestep(in);
	step++;

	printf("fish step t=%le V=%le current=%le %le\n",in->time,in->Vapplied,get_I(in),in->last_error);


	get_avg_np_pos(in,&nx,&px);
	inter_append(&fishoutn,in->time,nx);
	inter_append(&fishoutp,in->time,px);

	//FILE *out=fopen("./fish.dat","a");
	//fprintf(out,"%le %le %le\n",in->time,nx,px);
	//fclose(out);

	if (in->time>in->fish_stop) break;

	if (in->stop==TRUE)
	{
		break;

	}


}while(1);

in->go_time=FALSE;

frame_free(&frame);
char outpath[1000];
inter_sub(&fishoutn,fishoutn.data[0]);
inter_mod(&fishoutn);
inter_sub(&fishoutp,fishoutp.data[0]);
inter_mod(&fishoutp);
sprintf(outpath,"%s%s",in->outputpath,"fishn.dat");
inter_save(&fishoutn,outpath);


sprintf(outpath,"%s%s",in->outputpath,"fishp.dat");
inter_save(&fishoutp,outpath);

inter_add(&fishoutn,&fishoutp);

sprintf(outpath,"%s%s",in->outputpath,"fishnp.dat");
inter_save(&fishoutn,outpath);

inter_free(&fishoutn);
inter_free(&fishoutp);

}
//</clean>
