// General-purpose Photovoltaic Device Model gpvdm.com - a drift diffusion
// base/Shockley-Read-Hall model for 1st, 2nd and 3rd generation solarcells.
// The model can simulate OLEDs, Perovskite cells, and OFETs.
// 
// Copyright (C) 2008-2020 Roderick C. I. MacKenzie
// All rights reserved.
// 
// https://www.gpvdm.com
// r.c.i.mackenzie at googlemail.com
// 
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the GPVDM nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL Roderick C. I. MacKenzie BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "sim.h"
#include <math.h>
double min_pos_error=1e-14;


double get_p_error(struct device *in,double *b)
{
double tot=0.0;
int i;
for (i=0;i<in->ymeshpoints;i++)
{

	tot+=mod(b[i]);
}
return tot;
}


int solve_pos(struct device *in)
{
dprintf("Solve pos\n");
int i;

int N=in->ymeshpoints*3-2;


int M=in->ymeshpoints;
int *Ti =    malloc(N*sizeof(int));
int *Tj =    malloc(N*sizeof(int));
double *Tx = malloc(N*sizeof(double));
double *b = malloc(M*sizeof(double));


double phil;
double phic;
double phir;
double yl;
double yc;
double yr;
double dyl;
double dyr;
double dyc=0.0;
int ittr=0;
int pos=0;
double error=1000;
double el=0.0;
double ec=0.0;
double er=0.0;
double e0=0.0;
double e1=0.0;
int adv=FALSE;
int band;
	do
	{

		pos=0;
		for (i=0;i<in->ymeshpoints;i++)
		{


			if (i==0)
			{
				phil=in->Vl;
				el=in->epsilonr[0]*epsilon0;
				yl=in->ymesh[0]-(in->ymesh[1]-in->ymesh[0]);


			}else
			{
				el=in->epsilonr[i-1]*epsilon0;
				phil=in->phi[i-1];
				yl=in->ymesh[i-1];
			}

			if (i==(in->ymeshpoints-1))
			{
				phir=in->Vr;
				er=in->epsilonr[i]*epsilon0;
				yr=in->ymesh[i]+(in->ymesh[i]-in->ymesh[i-1]);
			}else
			{
				er=in->epsilonr[i+1]*epsilon0;
				phir=in->phi[i+1];
				yr=in->ymesh[i+1];

			}


			yc=in->ymesh[i];
			dyl=yc-yl;
			dyr=yr-yc;
			dyc=(dyl+dyr)/2.0;

			ec=in->epsilonr[i]*epsilon0;
			e0=(el+ec)/2.0;
			e1=(ec+er)/2.0;
			phic=in->phi[i];

			double dphidn=0.0;
			if (adv==FALSE)
			{
				dphidn=(q/(kb*in->Tl[i]))*in->Nc[i]*exp(((in->Fi[i]-(-in->phi[i]-in->Xi[i])))/(in->kTq));
			}else
			{
				dphidn=get_dn_den(in->Fi[i]-in->Ec[i],in->Tl[i],in->me[i],in->dostype[i]);
			}


			double dphidp=0.0;
			if (adv==FALSE)
			{
				dphidp=-(q/(kb*in->Tl[i]))*in->Nv[i]*exp((((-in->phi[i]-in->Xi[i]-in->Eg[i])-in->Fi[i]))/(in->kTq));
			}else
			{
				dphidp=-get_dp_den(in->xp[i]-in->tp[i],in->Tl[i],in->mh[i],in->dostype[i]);
			}
			double dphil=e0/dyl/dyc;
			double dphic=-(e0/dyl/dyc+e1/dyr/dyc);
			double dphir=e1/dyr/dyc;

			double dphi=dphil*phil+dphic*phic+dphir*phir;


			dphic+=-q*(dphidn-dphidp); //

			if (i!=0)
			{
				Ti[pos]=i;
				Tj[pos]=i-1;
				Tx[pos]=dphil;
				pos++;



			}

			Ti[pos]=i;
			Tj[pos]=i;
			Tx[pos]=dphic;



			pos++;



			if (i!=(in->ymeshpoints-1))
			{
				Ti[pos]=i;
				Tj[pos]=i+1;
				Tx[pos]=dphir;
				pos++;

			}

			b[i]=-(dphi-q*(in->n[i]-in->p[i])); //
			//in->n[i]=in->Nc[i]*exp(((in->Fi[i]-in->Ec[i])*q)/(kb*in->Tl[i]));


		}

		solver(M,N,Ti,Tj, Tx,b);


		for (i=0;i<in->ymeshpoints;i++)
		{
			in->phi[i]+=b[i]/(1.0+fabs(b[i]/in->posclamp/(300.0*kb/q)));
		}

		for (i=0;i<in->ymeshpoints;i++)
		{
			in->Ec[i]=-in->phi[i]-in->Xi[i];
			in->Ev[i]=-in->phi[i]-in->Xi[i]-in->Eg[i];

				if (adv==FALSE)
				{
					in->n[i]=in->Nc[i]*exp(((in->Fi[i]-in->Ec[i])*q)/(kb*in->Tl[i]));
					in->dn[i]=(q/(kb*in->Tl[i]))*in->Nc[i]*exp((q*(in->Fi[i]-in->Ec[i]))/(kb*in->Tl[i]));
				}else
				{
					in->n[i]=get_n_den(in->Fi[i]-in->Ec[i],in->Tl[i],in->me[i],in->dostype[i]);
					in->dn[i]=get_dn_den(in->Fi[i]-in->Ec[i],in->Tl[i],in->me[i],in->dostype[i]);
				}

				in->Fn[i]=in->Fi[i];
				in->Fp[i]=in->Fi[i];
				in->x[i]=in->phi[i]+in->Fn[i];
				in->t[i]=in->Xi[i];

				in->xp[i]=-(in->phi[i]+in->Fp[i]);
				in->tp[i]=in->Xi[i]+in->Eg[i];

				if (adv==FALSE)
				{
					in->p[i]=in->Nv[i]*exp(((in->xp[i]-in->tp[i])*q)/(kb*in->Tl[i]));
					in->dp[i]=(q/(kb*in->Tl[i]))*in->Nv[i]*exp(((in->xp[i]-in->tp[i])*q)/(kb*in->Tl[i]));
				}else
				{
					in->p[i]=get_p_den(in->xp[i]-in->tp[i],in->Tl[i],in->mh[i],in->dostype[i]);
					in->dp[i]=get_dp_den(in->xp[i]-in->tp[i],in->Tl[i],in->mh[i],in->dostype[i]);
				}


			for (band=0;band<in->srh_bands;band++)
			{
			in->Fnt[i][band]=-in->phi[i]-in->Xi[i]+dos_srh_get_fermi_n(in->n[i], in->p[i],band);
			in->xt[i][band]=in->phi[i]+in->Fnt[i][band];

			in->Fpt[i][band]=-in->phi[i]-in->Xi[i]-in->Eg[i]-dos_srh_get_fermi_p(in->n[i], in->p[i],band);
			in->xpt[i][band]=-(in->phi[i]+in->Fpt[i][band]);
			}

			in->tt[i]=in->Xi[i];
			in->tpt[i]=in->Xi[i]+in->Eg[i];

		}
		error=get_p_error(in,b);
		if (error<1)
		{
			adv=TRUE;
		}
		#ifdef dump_converge
		char temp[100];
		sprintf(temp,"%d Pos error = %e %d\n",ittr,error,adv);
		dprintf("%s",temp);
		#endif
		//printdebug( temp);



//double N=2.0*pow(((2.0*pi*kb*in->Tl[0]*in->me[0]*m0)/(hp*hp)),1.5);
//double test=N*exp((-3.000000e-03*q)/(kb*in->Tl[0]));
//printf("Check now %e %e\n",get_n_den(-3.000000e-03,in->Tl[0],in->me[0],in->dostype[0],dos_all),test);
//getchar();

//getchar();
		ittr++;
	}while((ittr<1000)&&(error>min_pos_error));


update_arrays(in);

	if (in->srh_sim==TRUE)
	{
/*		in->nfree[i]=get_n_den(in->Fi[i]-in->Ec[i],in->Tl[i],in->me[i],in->dostype[i]);
		in->pfree[i]=get_p_den(in->Ev[i]-in->Fi[i],in->Te[i],in->mh[i],in->dostype[i]);
		in->n[i]=get_n_den(in->Fi[i]-in->Ec[i],in->Tl[i],in->me[i],in->dostype[i]);
		in->p[i]=get_p_den(in->Ev[i]-in->Fi[i],in->Te[i],in->mh[i],in->dostype[i]);
		in->nlast[i]=in->n[i];
		in->plast[i]=in->p[i];
		in->dn[i]=get_dn_den(in->Fi[i]-in->Ec[i],in->Tl[i],in->me[i],in->dostype[i]);
		in->dp[i]=get_dp_den(in->Ev[i]-in->Fi[i],in->Te[i],in->mh[i],in->dostype[i]);
		in->mun[i]=get_n_mu(in->Fi[i]-in->Ec[i],in->Te[i],in->me[i],in->dostype[i]);
		in->mup[i]=get_p_mu(in->Ev[i]-in->Fi[i],in->Th[i],in->mh[i],in->dostype[i]);*/
		//<clean>
		time_init(in);
		//</clean>

	}



//solver_get_memory(in);
//fill_matrix(in);
//solver_free_memory(in);


//FILE *fermi_check=fopen("./fermi_check.dat","w");


/*double Ecl=0.0;
double Ecc=0.0;
double Ecr=0.0;

double wnl=0.0;
double wnc=0.0;
double wnr=0.0;
*/
//printf("%e %e %e %e\n",wnl,wnc,wnr,kb*300.0);
//getchar();

/*double nl=0.0;
double nc=0.0;
double nr=0.0;
double Dnl=0.0;
double Dnc=0.0;
double Dnr=0.0;


double xil=0.0;
double xir=0.0;
double Jnl=0.0;
double Jnr=0.0;
*/
//double J=0.0;

/*for (i=1;i<in->ymeshpoints-2;i++)
{
//double ddh=(in->ymesh[1]-in->ymesh[0]);
//double jntest=in->mun[i][0]*in->nfree[i]*(in->Ec[i+1]-in->Ec[i])/ddh+kb*in->Te[i]*in->mun[i][0]*(in->nfree[i+1]-in->nfree[i])/ddh;
printf("%e %e\n",in->mun[i][0],in->mup[i][0]);

Ecl=in->Ec[i-1];
Ecc=in->Ec[i];
Ecr=in->Ec[i+1];
//double wnl=(3.0/2.0)*kb*300.0;
//double wnc=(3.0/2.0)*kb*300.0;
//double wnr=(3.0/2.0)*kb*300.0;

/////////////////// free no einstein//////////////////////
wnl=q*(3.0/2.0)*kb*300.0;//in->n[i-1][0]/get_dn_den(in->Fi[i-1]-in->Ec[i-1],in->Tl[i-1],in->me[i-1],in->dostype[i-1],dos_all);
wnc=q*(3.0/2.0)*kb*300.0;// *in->n[i][0]/get_dn_den(in->Fi[i]-in->Ec[i],in->Tl[i],in->me[i],in->dostype[i],dos_all);
wnr=q*(3.0/2.0)*kb*300.0;// *in->n[i+1][0]/get_dn_den(in->Fi[i+1]-in->Ec[i+1],in->Tl[i+1],in->me[i+1],in->dostype[i+1],dos_all);

//printf("%e %e %e %e\n",wnl,wnc,wnr,kb*300.0);
//getchar();
nl=in->nfree[i-1];
nc=in->nfree[i];
nr=in->nfree[i+1];
Dnl=in->mun[i][0]*(2.0/3.0)*wnl/q;
Dnc=in->mun[i][0]*(2.0/3.0)*wnc/q;
Dnr=in->mun[i][0]*(2.0/3.0)*wnr/q;
Dnl=(Dnl+Dnc)/2.0;
Dnr=(Dnr+Dnc)/2.0;

xil=q*2.0*(3.0/2.0)*(Ecc-Ecl)/((wnc+wnl));
xir=q*2.0*(3.0/2.0)*(Ecr-Ecc)/((wnr+wnc));
Jnl=(Dnl/dyl)*(B(-xil)*nc-B(xil)*nl);
Jnr=(Dnr/dyr)*(B(-xir)*nr-B(xir)*nc);
double J_free_no_einstein=(Jnl+Jnr)*q/2.0;

///////////////////no einstein//////////////////////
wnl=q*(3.0/2.0)*kb*300.0;// in->n[i-1][0]/get_dn_den(in->Fi[i-1]-in->Ec[i-1],in->Tl[i-1],in->me[i-1],in->dostype[i-1],dos_all);
wnc=q*(3.0/2.0)*kb*300.0;// *in->n[i][0]/get_dn_den(in->Fi[i]-in->Ec[i],in->Tl[i],in->me[i],in->dostype[i],dos_all);
wnr=q*(3.0/2.0)*kb*300.0;// *in->n[i+1][0]/get_dn_den(in->Fi[i+1]-in->Ec[i+1],in->Tl[i+1],in->me[i+1],in->dostype[i+1],dos_all);

//printf("%e %e %e %e\n",wnl,wnc,wnr,kb*300.0);
//getchar();
nl=in->n[i-1][0];
nc=in->n[i][0];
nr=in->n[i+1][0];
Dnl=in->mun[i][0]*(2.0/3.0)*wnl/q;
Dnc=in->mun[i][0]*(2.0/3.0)*wnc/q;
Dnr=in->mun[i][0]*(2.0/3.0)*wnr/q;
Dnl=(Dnl+Dnc)/2.0;
Dnr=(Dnr+Dnc)/2.0;

xil=q*2.0*(3.0/2.0)*(Ecc-Ecl)/((wnc+wnl));
xir=q*2.0*(3.0/2.0)*(Ecr-Ecc)/((wnr+wnc));
Jnl=(Dnl/dyl)*(B(-xil)*nc-B(xil)*nl);
Jnr=(Dnr/dyr)*(B(-xir)*nr-B(xir)*nc);
double J_no_einstein=(Jnl+Jnr)*q/2.0;
//printf("test = %e %e\n",q*in->n[i][0]/get_dn_den(in->Fi[i]-in->Ec[i],in->Tl[i],in->me[i],in->dostype[i],dos_all),get_n_w(in->Fi[i]-in->Ec[i],in->Tl[i],in->me[i],in->dostype[i]));
//getchar();
///////////////////einstein//////////////////////
wnl=(3.0/2.0)*get_n_w(in->Fi[i-1]-in->Ec[i-1],in->Tl[i-1],in->me[i-1],in->dostype[i-1]);
wnc=(3.0/2.0)*get_n_w(in->Fi[i]-in->Ec[i],in->Tl[i],in->me[i],in->dostype[i]);
wnr=(3.0/2.0)*get_n_w(in->Fi[i+1]-in->Ec[i+1],in->Tl[i+1],in->me[i+1],in->dostype[i+1]);

//printf("%e %e %e %e\n",wnl,wnc,wnr,kb*300.0);
//getchar();
nl=in->n[i-1][0];
nc=in->n[i][0];
nr=in->n[i+1][0];
Dnl=in->mun[i][0]*(2.0/3.0)*wnl/q;
Dnc=in->mun[i][0]*(2.0/3.0)*wnc/q;
Dnr=in->mun[i][0]*(2.0/3.0)*wnr/q;
Dnl=(Dnl+Dnc)/2.0;
Dnr=(Dnr+Dnc)/2.0;

xil=q*2.0*(3.0/2.0)*(Ecc-Ecl)/((wnc+wnl));
xir=q*2.0*(3.0/2.0)*(Ecr-Ecc)/((wnr+wnc));
Jnl=(Dnl/dyl)*(B(-xil)*nc-B(xil)*nl);
Jnr=(Dnr/dyr)*(B(-xir)*nr-B(xir)*nc);
double J_einstein=(Jnl+Jnr)*q/2.0;

fprintf(fermi_check,"%e %e %e %e %e\n",in->ymesh[i],in->Jn[i][0],J_free_no_einstein,J_no_einstein,J_einstein);
}
fclose(fermi_check);
//printf("dumped\n");
getchar();*/


in->odes+=in->ymeshpoints;

#ifndef mindump
FILE *out;
out=fopena(in->outputpath,"./equ.dat","w");
for (i=0;i<in->ymeshpoints;i++)
{
	fprintf(out,"%e %e %e %e %e %e %e %e\n",in->ymesh[i],in->Ec[i],in->Ev[i],in->Fn[i],in->Fp[i],in->Fi[i],in->n[i],in->p[i]);
}
fclose(out);

out=fopena(in->outputpath,"./equ_phi.dat","w");
for (i=0;i<in->ymeshpoints;i++)
{
	fprintf(out,"%e %e\n",in->ymesh[i],in->phi[i]);
}
fclose(out);


out=fopena(in->outputpath,"./equ_np.dat","w");
for (i=0;i<in->ymeshpoints;i++)
{
	fprintf(out,"%e %e %e\n",in->ymesh[i],in->n[i],in->p[i]);
}
fclose(out);

out=fopena(in->outputpath,"./equ_np_trap.dat","w");
for (i=0;i<in->ymeshpoints;i++)
{
	fprintf(out,"%e ",in->ymesh[i]);
	for (band=0;band<in->srh_bands;band++)
	{
		fprintf(out,"%e %e ",in->nt[i][band],in->pt[i][band]);
	}
	fprintf(out,"\n");
}
fclose(out);

#endif
for (i=0;i<in->ymeshpoints;i++)
{
	in->nfinit[i]=in->n[i];
	in->pfinit[i]=in->p[i];
	in->ntinit[i]=0.0;//in->nt[i];
	in->ptinit[i]=0.0;//in->pt[i];
}

free(Ti);
free(Tj);
free(Tx);
free(b);
//printdebug("Done\n");
//getchar();
return 0;
}

